==> 0.4.1&file=in_ffsox-2a-0.4.1-src.tar.gz: in_ffsox-2a-0.4.1/in_ffsox/decoder.c <==
/*
 * decoder.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

#if AVCODEC_VER < 58 // [
#define AV_CODEC_ID_MJPEG CODEC_ID_MJPEG
#define AV_CODEC_ID_PNG CODEC_ID_PNG
#define AV_CODEC_ID_BMP CODEC_ID_BMP
#endif // ]

#define decoder_is_mjpeg(dec)  ( \
  AV_CODEC_ID_MJPEG==(dec)->t.ffi.vst->st->codec->codec_id \
  ||AV_CODEC_ID_PNG==(dec)->t.ffi.vst->st->codec->codec_id \
  ||AV_CODEC_ID_BMP==(dec)->t.ffi.vst->st->codec->codec_id \
)

#define decoder_is_video(dec) ( \
  0!=(dec)->ps->general.vid \
  &&NULL!=(dec)->v.mod \
  &&NULL!=(dec)->t.ffi.vst \
  &&!decoder_is_mjpeg(dec) \
)

#define decoder_overflow(dec) ( \
  queue_overflow(&(dec)->a.q) \
  &&(!(dec)->is_video||queue_overflow(&(dec)->v.q)) \
)

decoder_t *decoder_init(decoder_t *dec, prefs_t *ps, In_Module *mod)
{
  dec->ps=ps;
  dec->state=STATE_RUN;

  dec->hMutex=CreateMutex(
    NULL,   // _In_opt_  LPSECURITY_ATTRIBUTES lpMutexAttributes,
    FALSE,  // _In_      BOOL bInitialOwner,
    NULL    // _In_opt_  LPCTSTR lpName
  );

  if (NULL==dec->hMutex)
    goto mutex_error;

  dec->hEventWrite=CreateEvent(
    NULL,   // lpEventAttributes
    FALSE,  // bManualReset
    FALSE,  // bInitialState
    NULL    // lpName
  );

  if (NULL==dec->hEventWrite)
    goto event_error;

  if (NULL==audio_init(&dec->a,dec))
    goto audio_error;

  if (NULL==video_init(&dec->v,dec,mod))
    goto video_error;

  return dec;
video_error:
  audio_cleanup(&dec->a);
audio_error:
  CloseHandle(dec->hEventWrite);
event_error:
  CloseHandle(dec->hMutex);
mutex_error:
  return NULL;
}

decoder_t *decoder_cleanup(decoder_t *dec)
{
  video_cleanup(&dec->v);
  audio_cleanup(&dec->a);

  if (NULL!=dec->hEventWrite)
    CloseHandle(dec->hEventWrite);

  if (NULL!=dec->hMutex)
    CloseHandle(dec->hMutex);

  return dec;
}

#if 1 // {
double decoder_get_time(decoder_t *dec, In_Module *mod)
{
  Out_Module *outMod=mod->outMod;

  return dec->a.s.time
      +outMod->GetOutputTime()
	  -outMod->GetWrittenTime(); 
}
#endif // }

decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t nsam)
{
  track_t *t=&dec->t;

  if (NULL==track_init(t,fn,dec->ps,1))
    goto track_error;

  if (NULL==audio_realloc(&dec->a,t,nsam))
    goto audio_error;

  dec->is_video=decoder_is_video(dec);

  if (dec->is_video&&NULL==video_realloc(&dec->v,t))
    goto video_error;

  return dec;
video_error:
  audio_stop(&dec->a);
audio_error:
  track_cleanup(t);
track_error:
  return NULL;
}

void decoder_stop(decoder_t *dec, void (*killed)(void *), void *data)
{
  if (NULL!=killed)
    killed(data);

  if (dec->is_video)
    video_stop(&dec->v);

  audio_stop(&dec->a);
  track_cleanup(&dec->t);
}

void decoder_reset(decoder_t *dec)
{
  dec->state=STATE_RUN;
  audio_reset(&dec->a);

  if (dec->is_video)
    video_reset(&dec->v);
}

decoder_t *decoder_run(decoder_t *dec)
{
  decoder_reset(dec);

  dec->hThread=CreateThread(
    NULL,   // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,      // _In_       SIZE_T dwStackSize,
    decoder_thread,
            // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    dec,    // _In_opt_   LPVOID lpParameter,
    0,      // _In_       DWORD dwCreationFlags,
    NULL    // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==dec->hThread)
    goto error;

  return dec;
error:
  return NULL;
}

void decoder_pause(decoder_t *dec)
{
  WaitForSingleObject(dec->hMutex,INFINITE);

  if (STATE_RUN==dec->state) {
    dec->state=STATE_PAUSE;
	SetEvent(dec->a.q.hEventRead);

    if (dec->is_video)
	  SetEvent(dec->v.q.hEventRead);

	SetEvent(dec->hEventWrite);
  }
  
  ReleaseMutex(dec->hMutex);
}

void decoder_unpause(decoder_t *dec)
{
  WaitForSingleObject(dec->hMutex,INFINITE);

  if (STATE_PAUSE==dec->state) {
    dec->state=STATE_RUN;
	SetEvent(dec->a.q.hEventRead);

    if (dec->is_video)
	  SetEvent(dec->v.q.hEventRead);

	SetEvent(dec->hEventWrite);
  }
  
  ReleaseMutex(dec->hMutex);
}

void decoder_wait(decoder_t *dec)
{
  if (NULL!=dec->hThread) {
    WaitForSingleObject(dec->hThread,INFINITE);
    CloseHandle(dec->hThread);
    dec->hThread=NULL;
  }
}

void decoder_kill(decoder_t *dec)
{
  audio_kill(&dec->a);

  if (dec->is_video)
    video_kill(&dec->v);

  WaitForSingleObject(dec->hMutex,INFINITE);
  dec->state=STATE_KILL;
  SetEvent(dec->a.q.hEventRead);

  if (dec->is_video)
    SetEvent(dec->v.q.hEventRead);

  SetEvent(dec->hEventWrite);
  ReleaseMutex(dec->hMutex);
}

void decoder_eof(decoder_t *dec)
{
  WaitForSingleObject(dec->hMutex,INFINITE);

  if (STATE_RUN==dec->state) {
    dec->state=STATE_EOF;
    SetEvent(dec->a.q.hEventRead);

    if (dec->is_video)
      SetEvent(dec->v.q.hEventRead);

    SetEvent(dec->hEventWrite);
  }

  ReleaseMutex(dec->hMutex);
}

int decoder_seek(decoder_t *dec, int64_t pos_ms)
{
  ffi_t *ffi=&dec->t.ffi;
  // prefer then video stream for seeking.
  ffi_st_t *st=dec->is_video?ffi->vst:ffi->ast;
  int64_t seek_target=av_rescale_q(pos_ms*1000,AV_TIME_BASE_Q,
      st->st->time_base);
  int ret;

  decoder_kill(dec);
  decoder_wait(dec);

  avcodec_flush_buffers(ffi->ast->dc);

  if (dec->is_video)
    avcodec_flush_buffers(ffi->vst->dc);

  ret=avformat_seek_file(ffi->ic,st->index,INT64_MIN,seek_target,
        INT64_MAX,0);

  if (NULL==decoder_run(dec))
    goto error;

  return ret;
error:
  decoder_kill(dec);
  decoder_stop(dec,NULL,NULL);

  return -1;
}

void decoder_write(decoder_t *dec, queue_t *q, AVPacket *pkt)
{
  WaitForSingleObject(dec->hMutex,INFINITE);

  while (STATE_KILL!=dec->state&&decoder_overflow(dec)) {
    SignalObjectAndWait(dec->hMutex,dec->hEventWrite,INFINITE,FALSE);
    WaitForSingleObject(dec->hMutex,INFINITE);
  }

  if (STATE_KILL!=dec->state) {
    queue_write(q,pkt);
    SetEvent(q->hEventRead);
  }
  else
    av_free_packet(pkt);

  ReleaseMutex(dec->hMutex);
}

DWORD WINAPI decoder_thread(LPVOID lpParam)
{
  decoder_t *dec=lpParam;
  ffi_t *ffi=&dec->t.ffi;
  ffi_st_t *ast=ffi->ast,*vst=ffi->vst;
  queue_t *aq=&dec->a.q,*vq=&dec->v.q;
  AVPacket pkt;

  // initialize packet, set data to NULL, let the demuxer fill it
  ZERO(&pkt);
  av_init_packet(&pkt);

  // start sub-threads
  if (NULL==audio_run(&dec->a))
    goto cleanup_audio;

  if (dec->is_video&&NULL==video_run(&dec->v))
    goto cleanup_video;

  while (STATE_KILL!=dec->state&&0<=av_read_frame(ffi->ic,&pkt)) {
    if (pkt.stream_index==ast->index)
      decoder_write(dec,aq,&pkt);
    else if (dec->is_video&&pkt.stream_index==vst->index)
      decoder_write(dec,vq,&pkt);
    else
      av_free_packet(&pkt);
  }

  decoder_eof(dec);

  // wait for sub-threads.
  if (dec->is_video)
    video_wait(&dec->v);
cleanup_video:
  audio_wait(&dec->a);
cleanup_audio:
  return 0;
}

==> 0.4.2&file=in_ffsox-2a-0.4.2-src.tar.gz: in_ffsox-2a-0.4.2/in_ffsox/decoder.c <==
/*
 * decoder.c
 * Copyright (C) 2010 - 2013 Peter Belkner <pbelkner@snafu.de>
 * 
 * This program is free software: you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation, either version 3 of the License, or
 * (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program.  If not, see <http://www.gnu.org/licenses/>.
 */
#include "plugin.h"

#if AVCODEC_VER < 58 // [
#define AV_CODEC_ID_MJPEG CODEC_ID_MJPEG
#define AV_CODEC_ID_PNG CODEC_ID_PNG
#define AV_CODEC_ID_BMP CODEC_ID_BMP
#endif // ]

#define decoder_is_mjpeg(dec)  ( \
  AV_CODEC_ID_MJPEG==(dec)->t.ffi.vst->st->codec->codec_id \
  ||AV_CODEC_ID_PNG==(dec)->t.ffi.vst->st->codec->codec_id \
  ||AV_CODEC_ID_BMP==(dec)->t.ffi.vst->st->codec->codec_id \
)

#if defined (VIS_DELAY) // [
#define decoder_is_video(dec) ( \
  0!=(dec)->ps->general.vid.enabled \
  &&NULL!=(dec)->v.mod \
  &&NULL!=(dec)->t.ffi.vst \
  &&!decoder_is_mjpeg(dec) \
)
#else // ] [
#define decoder_is_video(dec) ( \
  0!=(dec)->ps->general.vid \
  &&NULL!=(dec)->v.mod \
  &&NULL!=(dec)->t.ffi.vst \
  &&!decoder_is_mjpeg(dec) \
)
#endif // ]

#define decoder_overflow(dec) ( \
  queue_overflow(&(dec)->a.q) \
  &&(!(dec)->is_video||queue_overflow(&(dec)->v.q)) \
)

decoder_t *decoder_init(decoder_t *dec, prefs_t *ps, In_Module *mod)
{
  dec->ps=ps;
  dec->state=STATE_RUN;

  dec->hMutex=CreateMutex(
    NULL,   // _In_opt_  LPSECURITY_ATTRIBUTES lpMutexAttributes,
    FALSE,  // _In_      BOOL bInitialOwner,
    NULL    // _In_opt_  LPCTSTR lpName
  );

  if (NULL==dec->hMutex)
    goto mutex_error;

  dec->hEventWrite=CreateEvent(
    NULL,   // lpEventAttributes
    FALSE,  // bManualReset
    FALSE,  // bInitialState
    NULL    // lpName
  );

  if (NULL==dec->hEventWrite)
    goto event_error;

  if (NULL==audio_init(&dec->a,dec))
    goto audio_error;

  if (NULL==video_init(&dec->v,dec,mod))
    goto video_error;

  return dec;
video_error:
  audio_cleanup(&dec->a);
audio_error:
  CloseHandle(dec->hEventWrite);
event_error:
  CloseHandle(dec->hMutex);
mutex_error:
  return NULL;
}

decoder_t *decoder_cleanup(decoder_t *dec)
{
  video_cleanup(&dec->v);
  audio_cleanup(&dec->a);

  if (NULL!=dec->hEventWrite)
    CloseHandle(dec->hEventWrite);

  if (NULL!=dec->hMutex)
    CloseHandle(dec->hMutex);

  return dec;
}

#if 1 // {
double decoder_get_time(decoder_t *dec, In_Module *mod)
{
  Out_Module *outMod=mod->outMod;

  return dec->a.s.time
      +outMod->GetOutputTime()
	  -outMod->GetWrittenTime(); 
}
#endif // }

decoder_t *decoder_start(decoder_t *dec, const in_char *fn, size_t nsam)
{
  track_t *t=&dec->t;

  if (NULL==track_init(t,fn,dec->ps,1)) {
    DMESSAGE("track_init() failed");
    goto track_error;
  }

  if (NULL==audio_realloc(&dec->a,t,nsam)) {
    DMESSAGE("audio_realloc() failed");
    goto audio_error;
  }

  dec->is_video=decoder_is_video(dec);

  if (dec->is_video&&NULL==video_realloc(&dec->v,t)) {
    DMESSAGE("video_realloc() failed");
    goto video_error;
  }

  return dec;
video_error:
  audio_stop(&dec->a);
audio_error:
  track_cleanup(t);
track_error:
  return NULL;
}

void decoder_stop(decoder_t *dec, void (*killed)(void *), void *data)
{
  if (NULL!=killed)
    killed(data);

  if (dec->is_video)
    video_stop(&dec->v);

  audio_stop(&dec->a);
  track_cleanup(&dec->t);
}

void decoder_reset(decoder_t *dec)
{
  dec->state=STATE_RUN;
  audio_reset(&dec->a);

  if (dec->is_video)
    video_reset(&dec->v);
}

decoder_t *decoder_run(decoder_t *dec)
{
  decoder_reset(dec);

  dec->hThread=CreateThread(
    NULL,   // _In_opt_   LPSECURITY_ATTRIBUTES lpThreadAttributes,
    0,      // _In_       SIZE_T dwStackSize,
    decoder_thread,
            // _In_       LPTHREAD_START_ROUTINE lpStartAddress,
    dec,    // _In_opt_   LPVOID lpParameter,
    0,      // _In_       DWORD dwCreationFlags,
    NULL    // _Out_opt_  LPDWORD lpThreadId
  );

  if (NULL==dec->hThread)
    goto error;

  return dec;
error:
  return NULL;
}

void decoder_pause(decoder_t *dec)
{
  WaitForSingleObject(dec->hMutex,INFINITE);

  if (STATE_RUN==dec->state) {
    dec->state=STATE_PAUSE;
	SetEvent(dec->a.q.hEventRead);

    if (dec->is_video)
	  SetEvent(dec->v.q.hEventRead);

	SetEvent(dec->hEventWrite);
  }
  
  ReleaseMutex(dec->hMutex);
}

void decoder_unpause(decoder_t *dec)
{
  WaitForSingleObject(dec->hMutex,INFINITE);

  if (STATE_PAUSE==dec->state) {
    dec->state=STATE_RUN;
	SetEvent(dec->a.q.hEventRead);

    if (dec->is_video)
	  SetEvent(dec->v.q.hEventRead);

	SetEvent(dec->hEventWrite);
  }
  
  ReleaseMutex(dec->hMutex);
}

void decoder_wait(decoder_t *dec)
{
  if (NULL!=dec->hThread) {
    WaitForSingleObject(dec->hThread,INFINITE);
    CloseHandle(dec->hThread);
    dec->hThread=NULL;
  }
}

void decoder_kill(decoder_t *dec)
{
  audio_kill(&dec->a);

  if (dec->is_video)
    video_kill(&dec->v);

  WaitForSingleObject(dec->hMutex,INFINITE);
  dec->state=STATE_KILL;
  SetEvent(dec->a.q.hEventRead);

  if (dec->is_video)
    SetEvent(dec->v.q.hEventRead);

  SetEvent(dec->hEventWrite);
  ReleaseMutex(dec->hMutex);
}

void decoder_eof(decoder_t *dec)
{
  WaitForSingleObject(dec->hMutex,INFINITE);

  if (STATE_RUN==dec->state) {
    dec->state=STATE_EOF;
    SetEvent(dec->a.q.hEventRead);

    if (dec->is_video)
      SetEvent(dec->v.q.hEventRead);

    SetEvent(dec->hEventWrite);
  }

  ReleaseMutex(dec->hMutex);
}

int decoder_seek(decoder_t *dec, int64_t pos_ms)
{
  ffi_t *ffi=&dec->t.ffi;
  // prefer then video stream for seeking.
  ffi_st_t *st=dec->is_video?ffi->vst:ffi->ast;
  int64_t seek_target=av_rescale_q(pos_ms*1000,AV_TIME_BASE_Q,
      st->st->time_base);
  int ret;

  decoder_kill(dec);
  decoder_wait(dec);

  avcodec_flush_buffers(ffi->ast->dc);

  if (dec->is_video)
    avcodec_flush_buffers(ffi->vst->dc);

  ret=avformat_seek_file(ffi->ic,st->index,INT64_MIN,seek_target,
        INT64_MAX,0);

  if (NULL==decoder_run(dec))
    goto error;

  return ret;
error:
  decoder_kill(dec);
  decoder_stop(dec,NULL,NULL);

  return -1;
}

void decoder_write(decoder_t *dec, queue_t *q, AVPacket *pkt)
{
  WaitForSingleObject(dec->hMutex,INFINITE);

  while (STATE_KILL!=dec->state&&decoder_overflow(dec)) {
    SignalObjectAndWait(dec->hMutex,dec->hEventWrite,INFINITE,FALSE);
    WaitForSingleObject(dec->hMutex,INFINITE);
  }

  if (STATE_KILL!=dec->state) {
    queue_write(q,pkt);
    SetEvent(q->hEventRead);
  }
  else
    av_free_packet(pkt);

  ReleaseMutex(dec->hMutex);
}

DWORD WINAPI decoder_thread(LPVOID lpParam)
{
  decoder_t *dec=lpParam;
  ffi_t *ffi=&dec->t.ffi;
  ffi_st_t *ast=ffi->ast,*vst=ffi->vst;
  queue_t *aq=&dec->a.q,*vq=&dec->v.q;
  AVPacket pkt;

  // initialize packet, set data to NULL, let the demuxer fill it
  ZERO(&pkt);
  av_init_packet(&pkt);

  // start sub-threads
  if (NULL==audio_run(&dec->a))
    goto cleanup_audio;

  if (dec->is_video&&NULL==video_run(&dec->v))
    goto cleanup_video;

  while (STATE_KILL!=dec->state&&0<=av_read_frame(ffi->ic,&pkt)) {
    if (pkt.stream_index==ast->index)
      decoder_write(dec,aq,&pkt);
    else if (dec->is_video&&pkt.stream_index==vst->index)
      decoder_write(dec,vq,&pkt);
    else
      av_free_packet(&pkt);
  }

  decoder_eof(dec);

  // wait for sub-threads.
  if (dec->is_video)
    video_wait(&dec->v);
cleanup_video:
  audio_wait(&dec->a);
cleanup_audio:
  return 0;
}
